"""Welcome to ADGTK (Agentic Data Generation Tool Kit). 

This is a pre-release development version of the project.
"""


__version__ = "0.2.0b1"

__author__ = "Fred Diehl"

__maintainer__ = "Fred Diehl"


# the "main items"
from .utils import get_scenario_logger, create_logger
