"""Contains the text for when no flag is set"""


intro = """
======================================
ADGTK - Agent Data Generation Tool-Kit
======================================

Interaction with ADGTK via the command line is done with the adgtk-mgr. It should be run from the root directory of a project folder or used to create a new project.

For more information please visit the project at http://adgtk.readthedocs.io/

"""
